open LexExpr

type expr =
  | Var of string
  | Int of int
  | Add of expr * expr
  | Mul of expr * expr

let token_string = function
  | ADD -> "+"
  | MUL -> "*"
  | LPAREN -> "("
  | RPAREN -> ")"
  | VAR x -> x
  | INT k -> string_of_int k
  | EOF -> "[EOF]"

let rec to_string = function
  | Var x -> x
  | Int k -> string_of_int k
  | Add (x, y) -> Printf.sprintf "(%s + %s)" (to_string x) (to_string y)
  | Mul (x, y) -> Printf.sprintf "(%s * %s)" (to_string x) (to_string y)


(*
  Original grammar:
  <expr>   ::= <term> + <expr> | <term>
  <term>   ::= <factor> * <term> | <factor>
  <factor> ::= <int> | <var> | (<expr>)

  This grammar is not LL(1), but it is after left-factoring:
  <expr>   ::= <term> <expr'>
  <expr'>  ::= + <expr> | epsilon
  <term>   ::= <factor> <term'>
  <term'>  ::= * <term> | epsilon
  <factor> ::= <int> | <var> | (<expr>)
*)
let parse buf =
  let lookahead = ref (lex buf) in

  (* Move to the next token *)
  let advance () =
    lookahead := lex buf
  in

  (* Read given token (or fail) then advance *)
  let eat token =
    if !lookahead == token then
      advance ()
    else
      failwith @@ Printf.sprintf "Expected %s, got %s"
        (token_string token)
        (token_string (!lookahead))
  in

  (* Create recursive functions, one for each non-terminal symbol. *)
  let rec expr : unit -> expr = fun () ->
    let t = term () in
    match expr' () with
    | Some e -> Add (t, e)
    | None -> t
  and expr' () =
    match !lookahead with
    | ADD ->
      advance ();
      Some (expr ())
    | _ -> None
  and term () =
    let f = factor () in
    match term' () with
    | Some e -> Mul (f, e)
    | None -> f
  and term' () =
    match !lookahead with
    | MUL ->
      advance ();
      Some (term ())
    | _ -> None
  and factor () =
    match !lookahead with
    | VAR x -> advance (); Var x
    | INT k -> advance (); Int k
    | LPAREN ->
      advance ();
      let e = expr () in
      eat RPAREN;
      e
    | _ ->
      failwith @@
      Printf.sprintf "Term: expecting VAR, INT, or expr; got %s"
        (token_string !lookahead)
  in
  expr ()

let () =
  let buf =
    Lexing.from_string (Sys.argv.(1))
  in
  Printf.printf "Successfully parsed: %s\n" (to_string (parse buf))
